import time
from locals import *

################################################################################

class AnimationFrame(object):
    def __init__(self, image, duration, hit_rect, hurt_rect = None, mirror_rects = False):
        self.image = image
        self.duration = duration # in seconds
        self.hit_rect = hit_rect

        if hurt_rect is None:
            self.hurt_rect = self.image.get_rect()
        else:
            self.hurt_rect = hurt_rect

        if mirror_rects:
            self.hit_rect = self._mirror_rect(self.hit_rect)
            self.hurt_rect = self._mirror_rect(self.hurt_rect)

        # for debugging, illustrate the hurt rect
#        pygame.draw.rect(self.image, (255, 0, 0), self.hurt_rect, 3)
#        if self.hit_rect is not None:
#            pygame.draw.rect(self.image, (0, 255, 0), self.hit_rect, 3)

    def _mirror_rect(self, r):
        if r is None:
            return None

        r.right = self.image.get_rect().width - r.left
        return r

    def copy(self):
        return AnimationFrame(self.image.copy(), self.duration, self.hit_rect, self.hurt_rect)

################################################################################

class Animation(object):
    def __init__(self, frames):
        self.frames = frames
        self.duration = sum([frame.duration for frame in frames])

        self.restart_anim()

    def restart_anim(self, notify_delegates = None):
        self.start_anim = time.time()
        if notify_delegates:
            if isinstance(notify_delegates, list):
                self.notify_delegates = notify_delegates
            elif callable(notify_delegates):
                self.notify_delegates = [notify_delegates]
        else:
            self.notify_delegates = []

    @property
    def cur_frame(self):
        cur_time = time.time()
        t = cur_time - self.start_anim

        if t > self.duration:
            self.notify_anim_done()
            t %= self.duration
            self.start_anim = cur_time - t

        running_duration = 0

        for frame in self.frames:
            running_duration += frame.duration
            if t <= running_duration:
                return frame

        # shouldn't get here, but just in case...
        return self.frames[-1]

    @property
    def image(self):
        return self.cur_frame.image

    @property
    def hit_rect(self):
         return self.cur_frame.hit_rect

    @property
    def hurt_rect(self):
        return self.cur_frame.hurt_rect

    def get_frame(self, frame_num):
        #return NoAnimation(self.frames[frame_num % len(self.frames)].image)
        return Animation([self.frames[frame_num % len(self.frames)]])

    def notify_anim_done(self):
        while self.notify_delegates:
            cb = self.notify_delegates.pop()
            cb()

    def overlay(self, animation):
        framelist = []
        for frame1, frame2 in zip(self.frames, animation.frames):
            frame1 = frame1.copy()
            frame1.image.blit(frame2.image, (0, 0))
            framelist.append(frame1)

        return Animation(framelist)

################################################################################

#class NoAnimation(object):
#    def __init__(self, image):
#        self.image = image
#
#    def restart_anim(self):
#        pass
#
#    def get_frame(self, frame_num):
#        return self
#
#    @property
#    def hurt_rect(self):
#        return self.image.get_rect()
def NoAnimation(image):
    return Animation([AnimationFrame(image, 1, None, None)])

################################################################################

# Animated: replaces "self.image" with "self.animation"
class Animated(object):
    @property
    def image(self):
        return self.animation.image

################################################################################

class AnimatedBlinkable(object):
    BLINK_FREQUENCY = 100 # msecs
    BLINK_ON_TIME = 60 # msecs

    BLINK_TIME = 1500 # msecs

    def __init__(self):
        self.blinking = False
        self._transparent_surface = pygame.Surface((0, 0))

    @property
    def image(self):
        if self.blinking and int(pygame.time.get_ticks() % self.BLINK_FREQUENCY) > self.BLINK_ON_TIME:
            return self._transparent_surface
        else:
            return self.animation.image

    def start_blinking(self):
        self.blink_stop_time = pygame.time.get_ticks() + self.BLINK_TIME
        self.blinking = True

    def update(self, *args):
        if self.blinking and pygame.time.get_ticks() >= self.blink_stop_time:
            self.blinking = False
