from animation import AnimatedBlinkable
from game import g
from locals import *
from scrollable_sprite import ScrollableSprite

class Character(ScrollableSprite, AnimatedBlinkable):
    INITIAL_JUMP_VELOC = 800    # pixels per second
    GRAVITY_ACCEL = 1200        # pixels per second^2

    ANIMATION_IDLE = None
    ANIMATION_JUMP = None

    def __init__(self, initial_x, initial_y, initial_facing):
        ScrollableSprite.__init__(self)
        AnimatedBlinkable.__init__(self)

        self.move_allowed = True
        self.can_fall = True

        self.left_pressed = False
        self.right_pressed = False
        self.dir_facing = initial_facing
        self.is_moving = False

        self.walk_speed_tweak = 0
        self.walk_accel_tweak = 0

        self.horiz_accel = 0
        self.horiz_veloc = 0

        self.vert_accel = 0
        self.vert_veloc = 0
        self.falling = False

        self.triggers_weak_platforms = False

        self.lowgrav = False

        self.walk_right_anim = g.assets[self.ANIMATION[DIR_RIGHT]]
        self.walk_left_anim = g.assets[self.ANIMATION[DIR_LEFT]]

        if self.ANIMATION_IDLE is None:
            self.idle_right_anim = self.walk_right_anim.get_frame(0)
            self.idle_left_anim = self.walk_left_anim.get_frame(0)
        else:
            self.idle_right_anim = g.assets[self.ANIMATION_IDLE[DIR_RIGHT]]
            self.idle_left_anim = g.assets[self.ANIMATION_IDLE[DIR_LEFT]]

        if self.ANIMATION_JUMP is None:
            self.jump_left_anim = self.idle_left_anim
            self.jump_right_anim = self.idle_right_anim
        else:
            self.jump_left_anim = g.assets[self.ANIMATION_JUMP[DIR_LEFT]]
            self.jump_right_anim = g.assets[self.ANIMATION_JUMP[DIR_RIGHT]]

        self.stand_anim()

        self.abs_rect = self.image.get_rect()
        self.pos = initial_x, initial_y

#        self.offset_by_attack = 0

    @property
    def hurt_rect(self):
        return self.animation.hurt_rect.move(*self.abs_rect.topleft)

    @property
    def initial_jump_veloc(self):
        if self.lowgrav:
            return self.INITIAL_JUMP_VELOC
        else:
            return self.INITIAL_JUMP_VELOC

    @property
    def gravity_accel(self):
        if self.lowgrav:
            return int(self.GRAVITY_ACCEL * 0.6)
        else:
            return self.GRAVITY_ACCEL

    @property
    def initial_walk_veloc(self):
        if self.lowgrav:
            return (self.INITIAL_WALK_VELOC + self.walk_speed_tweak) / 2
        else:
            return self.INITIAL_WALK_VELOC + self.walk_speed_tweak

    @property
    def walk_accel(self):
        if self.lowgrav:
            return (self.WALK_ACCEL + self.walk_accel_tweak) / 2
        else:
            return self.WALK_ACCEL + self.walk_accel_tweak

    @property
    def walk_accel_duration(self):
        if self.lowgrav:
            return self.WALK_ACCEL_DURATION
        else:
            return self.WALK_ACCEL_DURATION

    @property
    def pos(self):
        return self.fpos

    @pos.setter
    def pos(self, value):
        x, y = value
        self.fpos = float(x), float(y)

        #self.abs_rect.midbottom = x, y
        #self.hurt_rect.midbottom = x, y
        # I want to set self.hurt_rect.midbottom but I need to set self.abs_rect.midbottom...
        # Quick hack does this
        r = Rect(self.hurt_rect)
        a = self.abs_rect
        dx = r.left - a.left
        dy = r.top - a.top
        r.midbottom = x, y
        a.left = r.left - dx
        a.top = r.top - dy

    @property
    def falling_down(self):
        return self.falling and self.vert_veloc > 0

    def update(self, *args):
        self.update_position()
        ScrollableSprite.update(self, *args)
        AnimatedBlinkable.update(self, *args)

    def start_walk_right(self):
        self.right_pressed = True
        if not self.is_moving or self.dir_facing != DIR_RIGHT:
            self.dir_facing = DIR_RIGHT
            self.start_walk()

    def start_walk_left(self):
        self.left_pressed = True
        if not self.is_moving or self.dir_facing != DIR_LEFT:
            self.dir_facing = DIR_LEFT
            self.start_walk()

    def stop_walk_right(self):
        self.right_pressed = False
        if self.is_moving and self.dir_facing == DIR_RIGHT:
            if self.left_pressed:
                self.dir_facing = DIR_LEFT
                self.start_walk()
            else:
                self.stop_walk()

    def stop_walk_left(self):
        self.left_pressed = False
        if self.is_moving and self.dir_facing == DIR_LEFT:
            if self.right_pressed:
                self.dir_facing = DIR_RIGHT
                self.start_walk()
            else:
                self.stop_walk()

    def start_walk(self):
        if not self.move_allowed:
            return

        self.is_moving = True
        self.walk_start = pygame.time.get_ticks() / 1000. # in seconds...
        self.accel_time_remaining = self.walk_accel_duration
        if self.dir_facing == DIR_RIGHT:
            self.horiz_veloc = self.initial_walk_veloc
            self.horiz_accel = self.walk_accel
        elif self.dir_facing == DIR_LEFT:
            self.horiz_veloc = -self.initial_walk_veloc
            self.horiz_accel = -self.walk_accel

        self.update_anim()

    def stop_walk(self):
        self.horiz_accel = 0
        self.horiz_veloc = 0
        self.is_moving = False
        self.stand_anim()

    def stand_anim(self):
        if self.dir_facing == DIR_LEFT:
            self.animation = self.idle_left_anim
        else:
            self.animation = self.idle_right_anim

    def jump_anim(self):
        if self.dir_facing == DIR_LEFT:
            self.animation = self.jump_left_anim
        else:
            self.animation = self.jump_right_anim

    def walk_anim(self):
        if self.dir_facing == DIR_LEFT:
            self.animation = self.walk_left_anim
        else:
            self.animation = self.walk_right_anim

    def update_anim(self):
        if self.falling:
            self.jump_anim()
        elif self.is_moving:
            self.walk_anim()
        else:
            self.stand_anim()

    def resume_walk(self):
        # After punch animation for example. Resume walking in whatever direction we were in, only if needed
        if self.left_pressed or self.right_pressed:
            self.start_walk() # in direction facing...

    def jump(self):
        if self.move_allowed and not self.falling:
            self.start_falling()
            self.vert_veloc = -self.initial_jump_veloc

    def start_falling(self):
        if not self.can_fall:
            return

        self.falling = True
        self.fall_start = pygame.time.get_ticks() / 1000. # in seconds...
        self.vert_accel = self.gravity_accel
        self.vert_veloc = 0
        self.update_anim()

    def stop_falling(self):
        self.falling = False
        self.update_anim()

    def update_position(self):
        t = pygame.time.get_ticks() / 1000. # in seconds...

        x, y = self.pos

        # Are we moving horizontally?
        if self.is_moving:
            dt = t - self.walk_start
            self.walk_start = t # for next time

            if self.horiz_accel != 0 and dt > self.accel_time_remaining:
                # accelerate with dt = accel_time_remaining
                dt2 = self.accel_time_remaining
                # dx = v0 dt + (1/2) a dt dt
                # dv = a dt
                x += dt2 * (self.horiz_veloc + self.horiz_accel * dt2 / 2.)
                self.horiz_veloc += dt2 * self.horiz_accel

                # continue calculating without this period
                dt -= dt2

                # no more acceleration
                self.accel_time_remaining = 0
                self.horiz_accel = 0

            # dx = v0 dt + (1/2) a dt dt
            # dv = a dt
            x += dt * (self.horiz_veloc + self.horiz_accel * dt / 2.)
            self.horiz_veloc += dt * self.horiz_accel

            if self.accel_time_remaining > 0:
                self.accel_time_remaining -= dt

        # Are we free-falling? (y acceleration)
        if self.falling:
            dt = t - self.fall_start
            self.fall_start = t # for next time

            # dy = v0 dt + (1/2) a dt dt
            # dv = a dt
            y += dt * (self.vert_veloc + self.vert_accel * dt / 2.)
            self.vert_veloc += dt * self.vert_accel

        prev_rect = Rect(self.hurt_rect)
        self.pos = x, y
        rect = self.hurt_rect

        # Any collision at all with walls?
        idx = self.hurt_rect.collidelist([plat.abs_rect for plat in g.level.platforms if not plat.walk_through])
        if idx != -1:
            self.walked_into_wall()

        # Are we hitting solid platforms?
        colliding_platforms = [plat for plat in g.level.platforms
                               if plat.is_solid and self.hurt_rect.colliderect(plat.abs_rect)]
        while colliding_platforms:
            plat = colliding_platforms[0]
            plat_rect = plat.abs_rect

            x, y = self.pos
            dx, dy = 0, 0

            # Did we hit it from the right, left, top, or bottom?
            if plat_rect.left < rect.left < plat_rect.right <= prev_rect.left:
                # From the right
                if plat.walk_through:
                    colliding_platforms.remove(plat)
                else:
                    dx = plat_rect.right - rect.left
                    #self.abs_rect.left = plat_rect.right

            elif prev_rect.right <= plat_rect.left < rect.right < plat_rect.right:
                # From the left
                if plat.walk_through:
                    colliding_platforms.remove(plat)
                else:
                    dx = plat_rect.left - rect.right
                    #self.abs_rect.right = plat_rect.left

            elif prev_rect.bottom <= plat_rect.top < rect.bottom < plat_rect.bottom:
                # From the top (fell on this platform) - don't care if it's walk-through, solid is enough to stop falling
                self.stop_falling()
                dy = plat_rect.top - rect.bottom
                #self.abs_rect.bottom = plat_rect.top

            elif plat_rect.top < rect.top < plat_rect.bottom <= prev_rect.top:
                # From the bottom (jumped into this platform)
                if plat.walk_through:
                    colliding_platforms.remove(plat)
                else:
                    self.vert_veloc = 0 # "stop in midair"
                    dy = plat_rect.bottom - rect.top
                    #self.abs_rect.top = plat_rect.bottom

            else:
                colliding_platforms.remove(plat)

            if (dx, dy) != (0, 0):
                self.pos = x + dx, y + dy
                colliding_platforms = [plat for plat in g.level.platforms
                                       if plat.is_solid and self.hurt_rect.colliderect(plat.abs_rect)]

        # Fall-related checks:
        if self.can_fall and not self.falling:
            # Warn that we're about to fall.
            if self.is_moving:
                check_rect = g.collision_rect(self.hurt_rect)
                check_rect.bottom = self.hurt_rect.bottom

                if self.dir_facing == DIR_RIGHT:
                    pt_x, pt_y = (check_rect.right + 2, check_rect.bottom + 2)
                else:
                    pt_x, pt_y = (check_rect.left - 2, check_rect.bottom + 2)

                idx = Rect(pt_x, pt_y, 1, 1).collidelist([plat.abs_rect for plat in g.level.platforms if plat.is_solid])
                if idx == -1:
                    self.about_to_fall()

            # Check out whether we need to fall. If we don't have a solid platform underneath, fall
            underline = Rect(self.hurt_rect.left, self.hurt_rect.bottom + 1, self.hurt_rect.width, 1)
            solid_plats = [ plat for plat in g.level.platforms if plat.is_solid ]
            idx = underline.collidelist([ plat.abs_rect for plat in solid_plats ])

            if idx == -1:
                # Nothing solid underneath
                self.start_falling()
            else:
                # If the platform happens to be weak, trigger it
                plat = solid_plats[idx]
                if self.triggers_weak_platforms:
                    plat.step()

    def walked_into_wall(self):
        pass

    def about_to_fall(self):
        pass
